#define _POSIX_C_SOURCE 199309L
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <cal.h>
#include <calcl.h>
#include <stdbool.h>
#include <stdint.h>
#include "common.h"

typedef struct	gpu_state_s
{
    CALdevice		device;
    CALcontext		ctx;
    CALresource		constRes;
    CALresource		ihvRes;
    CALresource		sineRes;
    CALresource		outputRes;
    CALmem		constMem;
    CALmem		ihvMem;
    CALmem		sineMem;
    CALmem		outputMem;
    CALmodule		module;
    CALfunc		entry;
    CALprogramGrid	pg;
    CALevent		e;
}		gpu_state_t;

void fatal(const char *func_name)
{
    const char *cal_msg = calGetErrorString();
    const char *comp_msg = calclGetErrorString();
    fprintf(stderr, "%s failed - cal:[%s] calcl:[%s]\n",
	    func_name, cal_msg, comp_msg);
    // calGetErrorString error messages are prematurely truncated because of
    // stray NUL chars. print the message until 3 consecutive NUL chars are
    // encountered
    fprintf(stderr, "Full CAL error: ");
    for (int i = 0; i < 128; i++) {
	if (i > 1 && !cal_msg[i - 2] && !cal_msg[i - 1] && !cal_msg[i])
	    break;
	if (cal_msg[i])
	    fprintf(stderr, "%c", cal_msg[i]);
    }
    fprintf(stderr, "\n");
    exit(1);
}

int get_next_rand(void)
{
    static int val;
    static int initialized;
    if (!initialized)
      {
	val = rand();
	initialized = 1;
	return val;
      }
    return ++val;
}

void init_seed(const char *seed)
{
    unsigned t;
    if (!seed)
      {
	t = (unsigned)time(NULL);
	if ((unsigned)-1 == t)
	    perror("time"), exit(1);
      }
    else
	t = strtoul(seed, NULL, 0);
    FILE *f;
    const char *fname = ".seed";
    if (!(f = fopen(fname, "a")))
	perror(fname), exit(1);
    fprintf(f, "%u\n", t);
    if (fclose(f))
	perror("close"), exit(1);
    srand(t);
}

void show_ver()
{
    CALuint major, minor, imp;
    if (CAL_RESULT_OK != calGetVersion(&major, &minor, &imp))
	fatal("calGetVersion");
    printf("CAL version %u.%u.%u\n", major, minor, imp);
}

void show_attribs(CALuint devi, CALdeviceattribs *a)
{
    printf(
	    "=== device %u ===\n"
	    "target            %u\n"
	    "localRAM          %u MB\n"
	    "uncachedRemoteRAM %u MB\n"
	    "cachedRemoteRAM   %u MB\n"
	    "engineClock       %u MHz\n"
	    "memoryClock       %u MHz\n"
	    "wavefrontSize     %u\n"
	    "numberOfSIMD      %u\n"
	    "doublePrecision   %u\n"
	    "localDataShare    %u\n"
	    "globalDataShare   %u\n"
	    "globalGPR         %u\n"
	    "computeShader     %u\n"
	    "memExport         %u\n"
	    "pitch_alignment   %u\n"
	    "surface_alignment %u\n"
	    , devi, a->target, a->localRAM, a->uncachedRemoteRAM, a->cachedRemoteRAM,
	    a->engineClock, a->memoryClock, a->wavefrontSize, a->numberOfSIMD,
	    a->doublePrecision, a->localDataShare, a->globalDataShare,
	    a->globalGPR, a->computeShader, a->memExport, a->pitch_alignment,
	    a->surface_alignment);
}

void show_stats(struct timeval *v0, struct timeval *v1)
{
    long long ms0 = v0->tv_sec * 1000 + v0->tv_usec / 1000;
    long long ms1 = v1->tv_sec * 1000 + v1->tv_usec / 1000;
    printf("Execution time: %lld ms\n", ms1 - ms0);
#if 0
    printf("Achieved performance: %lld GFLOPS\n",
	    8 /* float op per MAD */ * (long long)NR_MAD_INSN *
	    NR_ITERATIONS * THREADS_PER_GRP * 10 / (long long)1e6 / (ms1 - ms0));
#endif
}

void copy_sine_values(CALuint *sinePtr, unsigned nr_values)
{
    CALuint a[] = {
	0xd76aa478, 0xe8c7b756, 0x242070db, 0xc1bdceee,
	0xf57c0faf, 0x4787c62a, 0xa8304613, 0xfd469501,
	0x698098d8, 0x8b44f7af, 0xffff5bb1, 0x895cd7be,
	0x6b901122, 0xfd987193, 0xa679438e, 0x49b40821,
	0xf61e2562, 0xc040b340, 0x265e5a51, 0xe9b6c7aa,
	0xd62f105d,  0x2441453, 0xd8a1e681, 0xe7d3fbc8,
	0x21e1cde6, 0xc33707d6, 0xf4d50d87, 0x455a14ed,
	0xa9e3e905, 0xfcefa3f8, 0x676f02d9, 0x8d2a4c8a,
	0xfffa3942, 0x8771f681, 0x6d9d6122, 0xfde5380c,
	0xa4beea44, 0x4bdecfa9, 0xf6bb4b60, 0xbebfbc70,
	0x289b7ec6, 0xeaa127fa, 0xd4ef3085,  0x4881d05,
	0xd9d4d039, 0xe6db99e5, 0x1fa27cf8, 0xc4ac5665,
	0xf4292244, 0x432aff97, 0xab9423a7, 0xfc93a039,
	0x655b59c3, 0x8f0ccc92, 0xffeff47d, 0x85845dd1,
	0x6fa87e4f, 0xfe2ce6e0, 0xa3014314, 0x4e0811a1,
	0xf7537e82, 0xbd3af235, 0x2ad7d2bb, 0xeb86d391,
    };
    if (nr_values != sizeof (a) / sizeof (*a))
	fprintf(stderr, "invalid nr_values: %u\n", nr_values), exit(1);
    for (unsigned i = 0; i < nr_values; i++)
	sinePtr[i] = a[i];
}

void prepare_run(CALuint devi, const unsigned char *bd_block, t_ihv *ihvs,
	gpu_state_t *gs, CALimage img)
{
    unsigned i;
    CALuint pitch = 0;
    // open device
    if (CAL_RESULT_OK != calDeviceOpen(&gs->device, devi))
	fatal("calDeviceOpen");
    if (CAL_RESULT_OK != calCtxCreate(&gs->ctx, gs->device))
	fatal("calCtxCreate");

    // message blocks (CAL doesn't allow UBYTE_1 for constant buffers)
    if (CAL_RESULT_OK != calResAllocLocal1D(&gs->constRes, gs->device,
		DATA_BYTES / 4, CAL_FORMAT_UINT_1, 0))
	fatal("calResAllocLocal1D (message blocks)");
    CALubyte *constPtr = 0;
    if (CAL_RESULT_OK != calResMap((CALvoid**)&constPtr, &pitch, gs->constRes, 0))
	fatal("calResMap");
    memcpy(constPtr, bd_block, DATA_BYTES);
    if (CAL_RESULT_OK != calResUnmap(gs->constRes))
	fatal("calResUnmap");
    if (CAL_RESULT_OK != calCtxGetMem(&gs->constMem, gs->ctx, gs->constRes))
	fatal("calCtxGetMem");

    // IHVs (CAL doesn't allow UBYTE_1 for constant buffers)
    if (CAL_RESULT_OK != calResAllocLocal1D(&gs->ihvRes, gs->device,
		IHV_BYTES / 4, CAL_FORMAT_UINT_1, 0))
	fatal("calResAllocLocal1D (IHVs)");
    CALuint *ihvPtr = 0;
    if (CAL_RESULT_OK != calResMap((CALvoid**)&ihvPtr, &pitch, gs->ihvRes, 0))
	fatal("calResMap");
    memcpy(ihvPtr, ihvs, 2 * sizeof (*ihvs));
    if (CAL_RESULT_OK != calResUnmap(gs->ihvRes))
	fatal("calResUnmap");
    if (CAL_RESULT_OK != calCtxGetMem(&gs->ihvMem, gs->ctx, gs->ihvRes))
	fatal("calCtxGetMem");

    // sine values (CAL doesn't allow UBYTE_1 for constant buffers)
    if (CAL_RESULT_OK != calResAllocLocal1D(&gs->sineRes, gs->device,
		SINE_BYTES / 4, CAL_FORMAT_UINT_1, 0))
	fatal("calResAllocLocal1D (sine buffer)");
    CALuint *sinePtr = 0;
    if (CAL_RESULT_OK != calResMap((CALvoid**)&sinePtr, &pitch, gs->sineRes, 0))
	fatal("calResMap");
    copy_sine_values(sinePtr, SINE_BYTES / 4);
    if (CAL_RESULT_OK != calResUnmap(gs->sineRes))
	fatal("calResUnmap");
    if (CAL_RESULT_OK != calCtxGetMem(&gs->sineMem, gs->ctx, gs->sineRes))
	fatal("calCtxGetMem");

    // global buffer
    /* the width has to be a multiple of 64, and neither the width nor the
     * height can be too large (for R700, the limit is 8192)
     */
    CALuint width = OUTPUT_BYTES;
    CALuint height = 1;
    while (width > 8192)
      {
	if (width % 64)
	    fprintf(stderr, "width %u not multiple of 64\n", width), exit(1);
	width /= 64;
	height *= 64;
      }
    if (width % 64)
      {
	CALuint width_old = width;
	width = width / 64 * 64 + 64;
	printf("rounding width from %u to %u\n", width_old, width);
      }
    if (CAL_RESULT_OK != calResAllocLocal2D(&gs->outputRes, gs->device,
		width, height, CAL_FORMAT_UBYTE_1,
		CAL_RESALLOC_GLOBAL_BUFFER))
	fatal("calResAllocLocal2D (global buffer)");
    CALuint *outputPtr = 0;
    if (CAL_RESULT_OK !=
	    calResMap((CALvoid**)&outputPtr, &pitch, gs->outputRes, 0))
	fatal("calResMap");
    // init every byte to 0
    for (i = 0; i < OUTPUT_BYTES / 4; i++)
	outputPtr[i] = 0;
    /* (NR_GROUPS * THREADS_PER_GRP) threads are going to compute
     * (NR_GROUPS * THREADS_PER_GRP * 4) paths. For each path, init
     * the original and computed x to different values. */
    for (i = 0; i < NR_GROUPS * THREADS_PER_GRP * 4; i++)
      {
	outputPtr[(i / 4) * (PATH_RES_BYTES / 4) + (i % 4)] =
	    outputPtr[(i / 4) * (PATH_RES_BYTES / 4) + (i % 4) + 8] =
	    get_next_rand();
      }
    if (CAL_RESULT_OK != calResUnmap(gs->outputRes))
	fatal("calResUnmap");
    if (CAL_RESULT_OK != calCtxGetMem(&gs->outputMem, gs->ctx, gs->outputRes))
	fatal("calCtxGetMem");
    // load module, get entry point
    if (CAL_RESULT_OK != calModuleLoad(&gs->module, gs->ctx, img))
	fatal("calModuleLoad");
    if (CAL_RESULT_OK != calModuleGetEntry(&gs->entry, gs->ctx, gs->module,
		"main"))
	fatal("calModuleGetEntry");
}

void load_module_and_execute(gpu_state_t *gs_base, CALuint count)
{
    for (CALuint devi = 0; devi < count; devi++)
      {
	gpu_state_t *gs = gs_base + devi;
	CALname n = 0;
	// bind message blocks
	if (CAL_RESULT_OK != calModuleGetName(&n, gs->ctx, gs->module, "cb0"))
	    fatal("calModuleGetName");
	if (CAL_RESULT_OK != calCtxSetMem(gs->ctx, n, gs->constMem))
	    fatal("calCtxSetMem (message blocks)");
	// bind IHVs
	if (CAL_RESULT_OK != calModuleGetName(&n, gs->ctx, gs->module, "cb2"))
	    fatal("calModuleGetName");
	if (CAL_RESULT_OK != calCtxSetMem(gs->ctx, n, gs->ihvMem))
	    fatal("calCtxSetMem (IHVs)");
	// bind sine buffer
	if (CAL_RESULT_OK != calModuleGetName(&n, gs->ctx, gs->module, "cb1"))
	    fatal("calModuleGetName");
	if (CAL_RESULT_OK != calCtxSetMem(gs->ctx, n, gs->sineMem))
	    fatal("calCtxSetMem (sine)");
	// bind global buffer
	if (CAL_RESULT_OK != calModuleGetName(&n, gs->ctx, gs->module, "g[]"))
	    fatal("calModuleGetName");
	if (CAL_RESULT_OK != calCtxSetMem(gs->ctx, n, gs->outputMem))
	    fatal("calCtxSetMem (output)");
	// init program grid
	CALprogramGrid pg = {
	    .func = gs->entry,
	    .gridBlock = { .width = THREADS_PER_GRP, .height = 1, .depth = 1 },
	    .gridSize = { .width = NR_GROUPS, .height = 1, .depth = 1 },
	    .flags = 0
	};
	gs->pg = pg;
	gs->e = 0;
      }
    // run
    struct timeval tv0, tv1;
    gettimeofday(&tv0, NULL);
    for (CALuint devi = 0; devi < count; devi++)
      {
	gpu_state_t *gs = gs_base + devi;
	//printf("Launching on device %u\n", devi);
	if (CAL_RESULT_OK != calCtxRunProgramGrid(&gs->e, gs->ctx, &gs->pg))
	    fatal("calCtxRunProgram");
      }
    // non-busy wait
    unsigned waiting_for = 0;
    if (count > sizeof (waiting_for) * 8)
	fprintf(stderr, "Cannot deal with %u devices\n", count), exit(1);
    for (unsigned i = 0; i < count; i++)
	waiting_for |= (1 << i);
    CALresult res;
    struct timespec req = { .tv_sec = 0, .tv_nsec = 1e6 };
    while (waiting_for)
      {
	for (CALuint devi = 0; devi < count; devi++)
	  {
	    gpu_state_t *gs = gs_base + devi;
	    if (!(waiting_for & (1 << devi)))
		continue;
	    res = calCtxIsEventDone(gs->ctx, gs->e);
	    if (res == CAL_RESULT_OK)
		waiting_for &= ~(1 << devi);
	    else if (res != CAL_RESULT_PENDING)
		fatal("calCtxIsEventDone");
	  }
	nanosleep(&req, NULL);
      }
    // show stats
    gettimeofday(&tv1, NULL);
    //show_stats(&tv0, &tv1);
}

void show_thread(int i, CALuint *out)
{
    printf("thread %i (%i %i %i %i iterations)\n"
	    "  start: x %08x %08x %08x %08x  y %08x %08x %08x %08x\n"
	    "    end: x %08x %08x %08x %08x  y %08x %08x %08x %08x\n",
	    i,
	    out[i * (PATH_RES_BYTES / 4) + 16],
	    out[i * (PATH_RES_BYTES / 4) + 17],
	    out[i * (PATH_RES_BYTES / 4) + 18],
	    out[i * (PATH_RES_BYTES / 4) + 19],
	    out[i * (PATH_RES_BYTES / 4) + 0],
	    out[i * (PATH_RES_BYTES / 4) + 1],
	    out[i * (PATH_RES_BYTES / 4) + 2],
	    out[i * (PATH_RES_BYTES / 4) + 3],
	    out[i * (PATH_RES_BYTES / 4) + 4],
	    out[i * (PATH_RES_BYTES / 4) + 5],
	    out[i * (PATH_RES_BYTES / 4) + 6],
	    out[i * (PATH_RES_BYTES / 4) + 7],
	    out[i * (PATH_RES_BYTES / 4) + 8],
	    out[i * (PATH_RES_BYTES / 4) + 9],
	    out[i * (PATH_RES_BYTES / 4) + 10],
	    out[i * (PATH_RES_BYTES / 4) + 11],
	    out[i * (PATH_RES_BYTES / 4) + 12],
	    out[i * (PATH_RES_BYTES / 4) + 13],
	    out[i * (PATH_RES_BYTES / 4) + 14],
	    out[i * (PATH_RES_BYTES / 4) + 15]);
}

void analyze_run(gpu_state_t *gs, unsigned nrrun)
{
    // read output
    CALuint *outputPtr = 0;
    CALuint pitch = 0;
    if (CAL_RESULT_OK != calResMap((CALvoid**)&outputPtr,
		&pitch, gs->outputRes, 0))
	fatal("calResMap");
    for (int i = 0; i < NR_GROUPS * THREADS_PER_GRP; i++)
      {
	(void)nrrun;
	bool dump_thread = false;
	// search for endpoints in the 4 pairs
	for (int pair = 0; pair < 4; pair++)
	    // XXX - hardcoded endpoint condition
	    if (!(0xfffffff0 & outputPtr[i * (PATH_RES_BYTES / 4) + 8 + pair]))
	      {
		printf("x,y start %08x,%08x end %08x,%08x (%i iterations) "
			"thread %i pair %i\n",
			outputPtr[i * (PATH_RES_BYTES / 4) + 0 + pair],
			outputPtr[i * (PATH_RES_BYTES / 4) + 4 + pair],
			outputPtr[i * (PATH_RES_BYTES / 4) + 8 + pair],
			outputPtr[i * (PATH_RES_BYTES / 4) + 12 + pair],
			outputPtr[i * (PATH_RES_BYTES / 4) + 16 + pair],
			i, pair);
		//dump_thread = true;
	      }
	if (dump_thread)
	    show_thread(i, outputPtr);
      }
    for (int i = 0; i < NR_GROUPS * THREADS_PER_GRP; i++)
	for (int pair = 0; pair < 4; pair++)
	    // XXX - hardcoded endpoint condition
	    if (!(0xfffffff0 & outputPtr[i * (PATH_RES_BYTES / 4) + 8 + pair]))
	      {
		// choose new x
		outputPtr[i * (PATH_RES_BYTES / 4) + 0 + pair] =
		    outputPtr[i * (PATH_RES_BYTES / 4) + 8 + pair] =
		    get_next_rand();
		// set y to 0
		outputPtr[i * (PATH_RES_BYTES / 4) + 4 + pair] =
		    outputPtr[i * (PATH_RES_BYTES / 4) + 12 + pair] =
		    0;
		// reset number of iterations !
		outputPtr[i * (PATH_RES_BYTES / 4) + 16 + pair] = 0;
	      }
    if (CAL_RESULT_OK != calResUnmap(gs->outputRes))
	fatal("calResUnmap");
}

void finish_run(gpu_state_t *gs)
{
    // unload module
    if (CAL_RESULT_OK != calModuleUnload(gs->ctx, gs->module))
	fatal("calModuleUnload");
    // release mem handle and deallocate message blocks
    if (CAL_RESULT_OK != calCtxReleaseMem(gs->ctx, gs->constMem))
	fatal("calCtxReleaseMem");
    if (CAL_RESULT_OK != calResFree(gs->constRes))
	fatal("calResFree");
    // release mem handle and deallocate IHVs
    if (CAL_RESULT_OK != calCtxReleaseMem(gs->ctx, gs->ihvMem))
	fatal("calCtxReleaseMem");
    if (CAL_RESULT_OK != calResFree(gs->ihvRes))
	fatal("calResFree");
    // release mem handle and deallocate sine buffer
    if (CAL_RESULT_OK != calCtxReleaseMem(gs->ctx, gs->sineMem))
	fatal("calCtxReleaseMem");
    if (CAL_RESULT_OK != calResFree(gs->sineRes))
	fatal("calResFree");
    // release mem handle and deallocate global buffer
    if (CAL_RESULT_OK != calCtxReleaseMem(gs->ctx, gs->outputMem))
	fatal("calCtxReleaseMem");
    if (CAL_RESULT_OK != calResFree(gs->outputRes))
	fatal("calResFree");
    // close device
    if (CAL_RESULT_OK != calCtxDestroy(gs->ctx))
	fatal("calCtxDestroy");
    if (CAL_RESULT_OK != calDeviceClose(gs->device))
	fatal("calDeviceClose");
}

void cal_puts(const CALchar *msg)
{
    fputs(msg, stdout);
}

void compile_and_run(CALuint count, const char *src, const unsigned char *blocks,
	t_ihv *ihvs, unsigned runs)
{
    CALuint devi;
    CALdeviceattribs attribs;
    // get attributes of the target devices
    for (devi = 0; devi < count; devi++)
      {
	attribs.struct_size = sizeof(CALdeviceattribs);
	if (CAL_RESULT_OK != calDeviceGetAttribs(&attribs, devi))
	    fatal("calDeviceGetAttribs");
	printf("Device %u: target=%u\n", devi, attribs.target);
	//show_attribs(devi, &attribs);
      }
    printf("Compiling for target=%u\n", attribs.target);
    // compile the object, link it into an image
    CALobject obj;
    if (CAL_RESULT_OK != calclCompile(&obj, CAL_LANGUAGE_IL, src, attribs.target))
	fatal("calclCompile");
    CALimage img;
    if (CAL_RESULT_OK != calclLink(&img, &obj, 1))
	fatal("calclLink");
    // disassemble the image
    //calclDisassembleImage(img, cal_puts);
    // prepare running it
    gpu_state_t *gs_base = malloc(count * sizeof (*gs_base));
    if (!gs_base)
	perror("malloc"), exit(1);
    for (CALuint devi = 0; devi < count; devi++)
	prepare_run(devi, blocks, ihvs, gs_base + devi, img);
    // load module and execute it
    for (unsigned nrrun = 0; nrrun < runs; nrrun++)
      {
	load_module_and_execute(gs_base, count);
	for (CALuint devi = 0; devi < count; devi++)
	    analyze_run(gs_base + devi, nrrun);
      }
    // finish running it
    for (CALuint devi = 0; devi < count; devi++)
	finish_run(gs_base + devi);
    // free the image and object
    if (CAL_RESULT_OK != calclFreeImage(img))
	fatal("calclFreeImage");
    if (CAL_RESULT_OK != calclFreeObject(obj))
	fatal("calclFreeObject");
}

off_t map_file(const char *fname, unsigned char **data)
{
    int fd;
    if (-1 == (fd = open(fname, O_RDONLY)))
	perror(fname), exit(1);
    struct stat st;
    if (-1 == fstat(fd, &st))
	perror("stat"), exit(1);
    if (MAP_FAILED == (*data = mmap(NULL, st.st_size, PROT_READ,
		    MAP_PRIVATE, fd, 0)))
	perror("mmap"), exit(1);
    return st.st_size;
}

void read_file_kernel(const char *fname, char **data)
{
    off_t s = map_file(fname, (unsigned char **)data);
    if (s < 2)
	fprintf(stderr, "file should be at least length 2: %s\n", fname),
	    exit(1);
    // either the last byte, or the next-to-last byte must be NUL
    if ((*data)[s - 1] && (*data)[s - 2])
	fprintf(stderr, "file is not NUL-terminated: %s\n", fname), exit(1);
}

void usage(const char *name)
{
    fprintf(stderr, "Usage: %s <file.il> <state-file> <runs> [<seed>]\n", name);
}

int main(int argc, char** argv)
{
    char *src = NULL;
    unsigned char *blocks = NULL;
    if (!(argc == 4 || argc == 5))
	usage(argv[0]), exit(1);
    t_ihv *ihvs = NULL;
    read_file_kernel(argv[1], &src);
    read_blocks_ihvs(&blocks, &ihvs, argv[2]);
    unsigned runs = strtoul(argv[3], NULL, 0);
    init_seed(argv[4]);
    if (CAL_RESULT_OK != calInit())
	fatal("calInit");
    setuid(42);
    show_ver();
    CALuint count;
    if (CAL_RESULT_OK != calDeviceGetCount(&count))
	fatal("calDeviceGetCount");
    printf("Found %u device%s\n", count, count > 1 ? "s" : "");
    if (count >= 1)
	compile_and_run(count, src, blocks, ihvs, runs);
    calShutdown();
    return 0;
}
