#!/usr/bin/perl -w
#
# MD5 implemented in AMD CAL IL.
#
# Author: Marc Bevand <bevand_m (at) epita.fr>
#

use strict;

# pick a strategy for implementing shifting: "a" or "b"
my $shift_strategy = "b";
my %map_shft_literals = ();
my $code = "";

# Returns word X[k] processed during the specified step (r7-r22).
sub step_to_x_k
{
    my ($step) = @_;
    my @elm = qw/xxxx yyyy zzzz wwww/;
    my $k;
    die "*bug*: step cannot be $step\n" if $step >= 64;
    if ($step < 16)      { $k = $step; }
    elsif ($step < 16*2) { $k = (5 * $step + 1) % 16; }
    elsif ($step < 16*3) { $k = (3 * $step + 5) % 16; }
    else                 { $k = (7 * $step) % 16; }
    return sprintf "r%i", 7 + $k;
}

# Returns the sine value T_i used in the specified step (cb1[?].?).
sub step_to_t_i
{
    my ($step) = @_;
    my @elm = qw/xxxx yyyy zzzz wwww/;
    die "*bug*: step cannot be $step\n" if $step >= 64;
    return sprintf "cb1[%i].%s", $step / 4, $elm[$step % 4];
}

# Returns the register corresponding to A,B,C,D (r0-r3).
sub abcd_to_r0123
{
    my %h = (A => "r0", B => "r1", C => "r2", D => "r3");
    for (my $i = 0; $i < scalar(@_); $i++) {
	die "*bug*: invalid variable: $_[$i]\n" if !exists($h{$_[$i]});
	$_[$i] = $h{$_[$i]};
    }
}

# Declares constant used by shifting operations.
sub declare_shft_literals
{
    my @literals = ();
    my @elm = qw/xxxx yyyy zzzz wwww/;
    for my $s (
	7, 12, 17, 22, # round 1
	5, 9, 14, 20,  # round 2
	4, 11, 16, 23, # round 3
	6, 10, 15, 21, # round 4
    ) {
	if ($shift_strategy eq 'a') {
	    push @literals, 32 - $s, $s;
	} elsif ($shift_strategy eq 'b') {
	    push @literals, 32 - $s, 2 ** $s;
	} else {
	    die "*bug*: invalid shift strategy: $shift_strategy\n";
	}
    }
    $code .= sprintf "  ; shift literals\n";
    for (my $i = 0; $i < scalar(@literals); $i++) {
	my $v = $literals[$i];
	my $lx = sprintf "l%d", $i / 4;
	my $e = $elm[$i % 4];
	$map_shft_literals{$v} = "$lx.$e";
	if (!($i % 4)) {
	    $code .= sprintf "  dcl_literal %s, %d", $lx, $v;
	} elsif (($i % 4) != 3) {
	    $code .= sprintf ", %d", $v;
	} else {
	    $code .= sprintf ", %d\n", $v;
	}
    }
}

# Returns the literal having the specified value.
sub get_shft_literal
{
    my ($v) = @_;
    exists($map_shft_literals{$v}) or die "cannot find literal $v\n";
    return $map_shft_literals{$v};
}

# There are 2 strategies to execute the rotate with 2 shifts:
# "A": uses 3 instructions (RIGHT SHIFT by "32-s", LEFT SHIFT by "s", OR the 2 results)
# "B": uses 2 instructions (RIGHT SHIFT by "32-s", MAD with "2**s" and "right shift result")
sub execute_shift_strategy
{
    my ($s_lit, $s_lit_compl, $s_lit_power) = @_;
    if ($shift_strategy eq 'a') {
	$code .= <<EOF;
    ishl r6, r5, $s_lit
    ushr r5, r5, $s_lit_compl
    ior  r5, r6, r5
EOF
    } else {
	$code .= <<EOF;
    ushr r6, r5, $s_lit_compl
    umad r5, r5, $s_lit_power, r6
EOF
    }
}

# Implements a round 1 step: t = x + ((t + F(x,y,z) + X[k] + T_i) <<< s)
sub round1_step
{
    my ($step, $t, $x, $y, $z, $s) = @_;
    my $x_k = step_to_x_k($step);
    my $t_i = step_to_t_i($step);
    abcd_to_r0123($t, $x, $y, $z);
    my $s_lit = get_shft_literal($s) if ($shift_strategy eq 'a');
    my $s_lit_compl = get_shft_literal(32 - $s);
    my $s_lit_power = get_shft_literal(2 ** $s) if ($shift_strategy eq 'b');
    $code .= "\n".
    "    ; step $step (round 1)\n".
    # F(x,y,z)
    "    ixor r5, $y, $z\n".
    "    and  r5, r5, $x\n".
    "    ixor r5, r5, $z\n".
    # X[k] + T_i
    "    iadd r6, $x_k, $t_i\n".
    # sum = t + F() + X[k] + T_i
    "    iadd r5, $t, r5\n".
    "    iadd r5, r5, r6\n";
    # sum <<< s
    execute_shift_strategy($s_lit, $s_lit_compl, $s_lit_power);
    # add x, store in t
    $code .= "    iadd $t, $x, r5\n";
}

# Implements a round 2 step: t = x + ((t + G(x,y,z) + X[k] + T_i) <<< s)
sub round2_step
{
    my ($step, $t, $x, $y, $z, $s) = @_;
    my $x_k = step_to_x_k($step);
    my $t_i = step_to_t_i($step);
    abcd_to_r0123($t, $x, $y, $z);
    my $s_lit = get_shft_literal($s) if ($shift_strategy eq 'a');
    my $s_lit_compl = get_shft_literal(32 - $s);
    my $s_lit_power = get_shft_literal(2 ** $s) if ($shift_strategy eq 'b');
    $code .= "\n".
    "    ; step $step (round 2)\n".
    # G(x,y,z)
    "    ixor r5, $x, $y\n".
    "    and  r5, r5, $z\n".
    "    ixor r5, r5, $y\n".
    # X[k] + T_i
    "    iadd r6, $x_k, $t_i\n".
    # sum = t + G() + X[k] + T_i
    "    iadd r5, $t, r5\n".
    "    iadd r5, r5, r6\n";
    # sum <<< s
    execute_shift_strategy($s_lit, $s_lit_compl, $s_lit_power);
    # add x, store in t
    $code .= "    iadd $t, $x, r5\n";
}

# Implements a round 3 step: t = x + ((t + H(x,y,z) + X[k] + T_i) <<< s)
sub round3_step
{
    my ($step, $t, $x, $y, $z, $s) = @_;
    my $x_k = step_to_x_k($step);
    my $t_i = step_to_t_i($step);
    abcd_to_r0123($t, $x, $y, $z);
    my $s_lit = get_shft_literal($s) if ($shift_strategy eq 'a');
    my $s_lit_compl = get_shft_literal(32 - $s);
    my $s_lit_power = get_shft_literal(2 ** $s) if ($shift_strategy eq 'b');
    $code .= "\n".
    "    ; step $step (round 3)\n".
    # H(x,y,z)
    "    ixor r5, $x, $y\n".
    "    ixor r5, r5, $z\n".
    # X[k] + T_i
    "    iadd r6, $x_k, $t_i\n".
    # sum = t + H() + X[k] + T_i
    "    iadd r5, $t, r5\n".
    "    iadd r5, r5, r6\n";
    # sum <<< s
    execute_shift_strategy($s_lit, $s_lit_compl, $s_lit_power);
    # add x, store in t
    $code .= "    iadd $t, $x, r5\n";
}

# Implements a round 4 step: t = x + ((t + I(x,y,z) + X[k] + T_i) <<< s)
sub round4_step
{
    my ($step, $t, $x, $y, $z, $s) = @_;
    my $x_k = step_to_x_k($step);
    my $t_i = step_to_t_i($step);
    abcd_to_r0123($t, $x, $y, $z);
    my $s_lit = get_shft_literal($s) if ($shift_strategy eq 'a');
    my $s_lit_compl = get_shft_literal(32 - $s);
    my $s_lit_power = get_shft_literal(2 ** $s) if ($shift_strategy eq 'b');
    $code .= "\n".
    "    ; step $step (round 4)\n".
    # I(x,y,z)
    "    inot r5, $z\n".
    "    ior r5, $x, r5\n".
    "    ixor r5, $y, r5\n".
    # X[k] + T_i
    "    iadd r6, $x_k, $t_i\n".
    # sum = t + I() + X[k] + T_i
    "    iadd r5, $t, r5\n".
    "    iadd r5, r5, r6\n";
    # sum <<< s
    execute_shift_strategy($s_lit, $s_lit_compl, $s_lit_power);
    # add x, store in t
    $code .= "    iadd $t, $x, r5\n";
}

my $output = shift;
open STDOUT,">$output" or die "can't open $output: $!";
$code .= <<EOF;
il_cs
  ; must be equal to THREADS_PER_GRP
  dcl_num_thread_per_group 192
  ; message blocks
  dcl_cb cb0[8]
  ; sine values
  dcl_cb cb1[16]
  ; IHV values
  dcl_cb cb2[2]
EOF
declare_shft_literals();
$code .= <<EOF;
  ; other constants
  ; l101.x is set to 4 in order to check the remaining 32-4=28 bits of x
  ; to determine when an endpoint is reached
  dcl_literal l100, 0x10000, -1, 1, 5
  dcl_literal l101, 4, 0, 0, 0

  ; r0-r3   A,B,C,D
  ; r4.x    counter
  ; r5-r6   temporary values
  ; r7-r20  X[0]-X[13]
  ; r21-r22 x,y (aka X[14],X[15])
  ; r23-r26 A,B,C,D values before processing the current block

  ; init x,y
  umul r5.x, vaTid.x, l100.w
  mov r21, g[r5.x+2]
  mov r22, g[r5.x+3]

  mov r4.x, l100.x
  whileloop
    ; break if counter is 0 or if an endpoint is found
    break_logicalz r4.x
    ; XXX - hardcoded endpoint condition
    ushr r5, r21, l101.x
    break_logicalz r5.x
    break_logicalz r5.y
    break_logicalz r5.z
    break_logicalz r5.w
    ; init X[0]-X[13] according to the parity of x
    and r5, r21, l100.z
    cmov_logical  r7, r5, cb0[0].x, cb0[4].x
    cmov_logical  r8, r5, cb0[0].y, cb0[4].y
    cmov_logical  r9, r5, cb0[0].z, cb0[4].z
    cmov_logical r10, r5, cb0[0].w, cb0[4].w
    cmov_logical r11, r5, cb0[1].x, cb0[5].x
    cmov_logical r12, r5, cb0[1].y, cb0[5].y
    cmov_logical r13, r5, cb0[1].z, cb0[5].z
    cmov_logical r14, r5, cb0[1].w, cb0[5].w
    cmov_logical r15, r5, cb0[2].x, cb0[6].x
    cmov_logical r16, r5, cb0[2].y, cb0[6].y
    cmov_logical r17, r5, cb0[2].z, cb0[6].z
    cmov_logical r18, r5, cb0[2].w, cb0[6].w
    cmov_logical r19, r5, cb0[3].x, cb0[7].x
    cmov_logical r20, r5, cb0[3].y, cb0[7].y
    ; init A,B,C,D according to the parity of x
    cmov_logical r0, r5, cb2[0].x, cb2[1].x
    cmov_logical r1, r5, cb2[0].y, cb2[1].y
    cmov_logical r2, r5, cb2[0].z, cb2[1].z
    cmov_logical r3, r5, cb2[0].w, cb2[1].w
    ; make copies of their initial values
    mov r23, r0
    mov r24, r1
    mov r25, r2
    mov r26, r3
EOF
my $s = 0;

round1_step($s++,'A','B','C','D', '7');
round1_step($s++,'D','A','B','C','12');
round1_step($s++,'C','D','A','B','17');
round1_step($s++,'B','C','D','A','22');
round1_step($s++,'A','B','C','D', '7');
round1_step($s++,'D','A','B','C','12');
round1_step($s++,'C','D','A','B','17');
round1_step($s++,'B','C','D','A','22');
round1_step($s++,'A','B','C','D', '7');
round1_step($s++,'D','A','B','C','12');
round1_step($s++,'C','D','A','B','17');
round1_step($s++,'B','C','D','A','22');
round1_step($s++,'A','B','C','D', '7');
round1_step($s++,'D','A','B','C','12');
round1_step($s++,'C','D','A','B','17');
round1_step($s++,'B','C','D','A','22');

round2_step($s++,'A','B','C','D', '5');
round2_step($s++,'D','A','B','C', '9');
round2_step($s++,'C','D','A','B','14');
round2_step($s++,'B','C','D','A','20');
round2_step($s++,'A','B','C','D', '5');
round2_step($s++,'D','A','B','C', '9');
round2_step($s++,'C','D','A','B','14');
round2_step($s++,'B','C','D','A','20');
round2_step($s++,'A','B','C','D', '5');
round2_step($s++,'D','A','B','C', '9');
round2_step($s++,'C','D','A','B','14');
round2_step($s++,'B','C','D','A','20');
round2_step($s++,'A','B','C','D', '5');
round2_step($s++,'D','A','B','C', '9');
round2_step($s++,'C','D','A','B','14');
round2_step($s++,'B','C','D','A','20');

round3_step($s++,'A','B','C','D', '4');
round3_step($s++,'D','A','B','C','11');
round3_step($s++,'C','D','A','B','16');
round3_step($s++,'B','C','D','A','23');
round3_step($s++,'A','B','C','D', '4');
round3_step($s++,'D','A','B','C','11');
round3_step($s++,'C','D','A','B','16');
round3_step($s++,'B','C','D','A','23');
round3_step($s++,'A','B','C','D', '4');
round3_step($s++,'D','A','B','C','11');
round3_step($s++,'C','D','A','B','16');
round3_step($s++,'B','C','D','A','23');
round3_step($s++,'A','B','C','D', '4');
round3_step($s++,'D','A','B','C','11');
round3_step($s++,'C','D','A','B','16');
round3_step($s++,'B','C','D','A','23');

round4_step($s++,'A','B','C','D', '6');
round4_step($s++,'D','A','B','C','10');
round4_step($s++,'C','D','A','B','15');
round4_step($s++,'B','C','D','A','21');
round4_step($s++,'A','B','C','D', '6');
round4_step($s++,'D','A','B','C','10');
round4_step($s++,'C','D','A','B','15');
round4_step($s++,'B','C','D','A','21');
round4_step($s++,'A','B','C','D', '6');
round4_step($s++,'D','A','B','C','10');
round4_step($s++,'C','D','A','B','15');
round4_step($s++,'B','C','D','A','21');
round4_step($s++,'A','B','C','D', '6');
round4_step($s++,'D','A','B','C','10');
round4_step($s++,'C','D','A','B','15');
round4_step($s++,'B','C','D','A','21');

$code .= <<EOF;

    ; finalize the computation of A,B,C,D
    iadd r0, r0, r23
    iadd r1, r1, r24
    iadd r2, r2, r25
    iadd r3, r3, r26
    ; compute x,y (x = A; y = C - D)
    mov r21, r0
    iadd r22, r2, r3_neg(xyzw)
    ; decrement loop counter
    iadd r4.x, r4.x, l100.y
  endloop

  umul r5.x, vaTid.x, l100.w
  mov g[r5.x+2], r21
  mov g[r5.x+3], r22
  ; increment the 4 iteration counters
  iadd r5.y, l100.x, r4_neg(x).x
  iadd g[r5.x+4], g[r5.x+4], r5.yyyy
end
\x00
EOF
print $code;

# eof
