#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <openssl/md5.h>
#include <glib.h>
#include <unistd.h>
#include "common.h"

void next_node(t_node *n, unsigned char *blocks, t_ihv *ihvs)
{
    // pick block/IHV according to the parity of x
    unsigned i = 1 - (n->x & 1);
    MD5_CTX c;
    MD5_Init(&c);
    c.A = ihvs[i].a;
    c.B = ihvs[i].b;
    c.C = ihvs[i].c;
    c.D = ihvs[i].d;
    MD5_Update(&c, blocks + i * 64, 64 - 8);
    MD5_Update(&c, &n->x, 4);
    MD5_Update(&c, &n->y, 4);
    // compute x,y (x = A; y = C - D)
    n->x = c.A;
    n->y = c.C - c.D;
    /*
    printf("A %08x\nB %08x\nC %08x\nD %08x\n",
	    swapu32(c.A), swapu32(c.B), swapu32(c.C), swapu32(c.D));
	    */
}

gboolean node_equal(t_node *a, t_node *b)
{
    return a->x == b->x && a->y == b->y;
}

gpointer node_to_ptr(t_node *n)
{
    // x is stored in the 32 LSBs, y in the 32 MSBs
    intptr_t p = (intptr_t)n->x | ((intptr_t)n->y << 32);
    return (gpointer)p;
}

t_node ptr_to_node(gpointer v)
{
    t_node res = {
	.x = (intptr_t)v,
	.y = (intptr_t)v >> 32,
    };
    return res;
}

/*
 * Search for the colliding nodes in 2 paths. Builds a hash table
 * containing all the nodes of the shortest table:
 * - key: current node
 * - value: previous node (can be NULL for first node)
 */
void search_coll(unsigned char *blocks, t_ihv *ihvs, t_node *end,
	t_node *starts, unsigned *iters)
{
    unsigned shortest_path = iters[0] < iters[1] ? 0 : 1;
    unsigned longest_path = 1 - shortest_path;
    // build table
    GHashTable *table = g_hash_table_new(NULL, NULL);
    t_node prev = { .x = 0, .y = 0 };
    t_node cur = starts[shortest_path];
    unsigned i;
    for (i = 0; i <= iters[shortest_path]; i++)
      {
	int last = i == iters[shortest_path];
	if (i)
	  {
	    prev = cur;
	    next_node(&cur, blocks, ihvs);
	  }
	if (!(i % 301747) || last || !cur.x)
	    printf("storing %08x,%08x -> %08x,%08x (%u/%u)\n",
		    prev.x, prev.y, cur.x, cur.y, i, iters[shortest_path]);
	g_hash_table_insert(table, node_to_ptr(&cur), node_to_ptr(&prev));
      }
    if (!node_equal(&cur, end))
	fprintf(stderr, "last node should have been %08x,%08x\n", end->x, end->y),
	    exit(1);
    // search it
    prev.x = prev.y = 0;
    cur = starts[longest_path];
    int match = 0;
    gpointer _prev2 = 0;
    for (i = 0; !match && i <= iters[longest_path]; i++)
      {
	int last = i == iters[longest_path];
	if (i)
	  {
	    prev = cur;
	    next_node(&cur, blocks, ihvs);
	  }
	if (g_hash_table_lookup_extended(table, node_to_ptr(&cur), NULL, &_prev2))
	    last = match = 1;
	if (!(i % 301747) || last)
	    printf("searching %08x,%08x -> %08x,%08x (%u/%u)\n",
		    prev.x, prev.y, cur.x, cur.y, i, iters[longest_path]);
      }
    if (!match)
	fprintf(stderr, "should have found the node, but didn't\n"), exit(1);
    t_node prev2 = ptr_to_node(_prev2);
    if ((!prev.x && !prev.y) || (!prev2.x && !prev2.y))
	printf("\"robin hood\" case:\n");
    else
	printf("collision found:\n");
    printf("  %08x,%08x -> %08x,%08x (short path)\n"
	    "  %08x,%08x -> %08x,%08x (long path)\n",
	    prev2.x, prev2.y, cur.x, cur.y,
	    prev.x, prev.y, cur.x, cur.y);
    // destroy it
    g_hash_table_destroy(table);
}

int main(int argc, char **argv)
{
    // this program is hackish and stores 2 32-bit values in gpointers (sorry!)
    if (sizeof (gpointer) != 8)
	fprintf(stderr, "sizeof (gpointer) == %zi not supported (should be 8)\n",
		sizeof (gpointer)), exit(1);
    if (argc != 7)
	fprintf(stderr,
		"Usage: %s <state> <x_end,y_end> <x1,y1> <iter1> <x2,y2> <iter2>\n",
		argv[0]), exit(1);
    unsigned char *blocks;
    t_ihv *ihvs;
    read_blocks_ihvs(&blocks, &ihvs, argv[1]);
    t_node end, start[2];
    unsigned iter[2];
    read_node(&end, argv[2]);
    read_node(start + 0, argv[3]);
    iter[0] = strtoul(argv[4], NULL, 0);
    read_node(start + 1, argv[5]);
    iter[1] = strtoul(argv[6], NULL, 0);
    search_coll(blocks, ihvs, &end, start, iter);
    return 0;
}
